//! this example demonstrates the type of four-core clocked inverter conceived
//! of in the https://uninsane.org/ blogpost.
//!
//! we're going for something like this:
//!
//! ```
//!          ____         ____         ____         ____
//!  + ___  /    \ +___  /    \ +___  /    \ +___  /    \  ___ +
//! IN ___ |o    o| ___ |x    o| ___ |x    o| ___ |x    o| ___ OUT
//!  -      \_o__/ -     \_o__/ -     \_o__/ -     \_o__/      -
//!           ||           ||           ||           ||
//!          +  -         +  -         +  -         +  -
//!          CTL0         CTL1         CTL2         CTL3
//! ```
//!
//! conventions:
//! - CW core polarization is logic '1'.
//! - differential signals are assumed to flow clockwise. i.e. left-to-right, with + on top and -
//!   on bottom.
//! - 'x' denotes a signal going "into" the page. 'o' for a signal going "out of" the page.
//! - the core driven by CTLn is named Sn, interchangeably used to represent the state (0 or 1) of
//!   that core
//!
//! hence:
//! - the input into core S0 has current flowing out of the page and tends to drive S0 to logic '0'.
//! - the input into S1, S2, S3 tends these cores to logic '1'.
//! - a positive current to CTLn will "clear" Sn to the logic '0' state.
//!
//! in this device, S0 is effectively an inverter, with S1, S2, S3 acting as buffers.
//! for the purpose of this simulation, all terminal wires are closed loops either explicitly
//! driven or measured.
//!
//! to run this, from toplevel directory:
//! ```
//! $ cargo run --release --bin multi_core_inverter
//! $ pushd crates/coremem; cargo run --release --bin viewer ../../out/applications/multi_core_inverter/0/ ; popd
//! ```

use coremem::geom::{Coord as _, Meters, Torus};
use coremem::mat::{Ferroxcube3R1MH, IsoConductorOr, IsomorphicConductor};
use coremem::meas;
use coremem::real::{self, Real as _};
use coremem::sim::spirv::{self, SpirvSim};
use coremem::sim::units::Seconds;
use coremem::stim::{CurlStimulus, Exp, Gated, Sinusoid1, TimeVarying as _};
use coremem::Driver;

type R = real::R32;
type Mat = IsoConductorOr<R, Ferroxcube3R1MH>;
// type Backend = spirv::CpuBackend;
type Backend = spirv::WgpuBackend;
type Sim = SpirvSim::<R, Mat, Backend>;

#[derive(Clone, Copy)]
enum DriveType {
    HoldHigh,
    ReleaseHigh,
    HoldLow,
    ReleaseLow,
    Float,
}

impl DriveType {
    fn direction(&self) -> i32 {
        use DriveType::*;
        match self {
            HoldHigh | ReleaseHigh => 1,
            HoldLow | ReleaseLow => -1,
            Float => 0,
        }
    }
    fn is_hold(&self) -> bool {
        match self {
            Self::HoldHigh | Self::HoldLow => true,
            _ => false,
        }
    }
    fn is_release(&self) -> bool {
        match self {
            Self::ReleaseHigh | Self::ReleaseLow => true,
            _ => false,
        }
    }
}

fn main() {
    coremem::init_logging();
    // coremem::init_debug();
    let um = |n| n as f32 * 1e-6;
    // let ns = |n| n as f32 * 1e-9;
    let ps = |n| n as f32 * 1e-12;
    let feat_size = um(10);

    let input_magnitude = 1.0e9;
    let clock_phase_duration = ps(2000);

    // 's' = core (ferromagnetic part)
    let s_major = um(160);
    let s_minor = um(30);
    // 'io' = drive/control wire
    let io_major = um(80);
    let io_minor = um(30);
    let coupling_major = um(130);
    let coupling_minor = um(30);

    // coords for core 'n'
    let sx = |n| um((n+1) * 400);
    let sy = um(400);
    let sz = um(280);
    // x-coord for the loop coupling core 'n' to core 'n+1'
    let couplingx = |n| sx(n) + um(200);

    let sim_bounds = |num_cores| Meters::new(sx(num_cores), sy * 2.0, sz * 2.0);
    let sim_padding = Meters::new(um(80), um(80), um(80));

    // core '0' gets an external input in place of its coupling loop
    let input0 = Torus::new_xz(Meters::new(sx(0) - s_major, sy, sz), io_major, io_minor);
    // the last core gets an external output in place of its coupling loop
    let sense = |n| Torus::new_xz(Meters::new(sx(n) + s_major, sy, sz), io_major, io_minor);
    // control loop for core n (alternately called "drive" loop)
    let ctl = |n| Torus::new_yz(Meters::new(sx(n), sy + s_major, sz), io_major, io_minor);
    let s = |n| Torus::new_xy(Meters::new(sx(n), sy, sz), s_major, s_minor);
    // coupling(n) is the wire which couples core n into core n+1
    let coupling = |n| Torus::new_xz(Meters::new(couplingx(n), sy, sz), coupling_major, coupling_minor);

    let control_signal = |driver: &mut Driver<Sim>, region: &Torus, cycle: u32, ty: DriveType| {
        let area = region.cross_section();
        let amp = ty.direction() as f32 * input_magnitude / area;
        let start = clock_phase_duration * cycle as f32;
        if ty.is_hold() {
            // simple square wave:
            let wave = Gated::new(amp, start, start + clock_phase_duration);
            let stim = CurlStimulus::new(
                region.clone(),
                wave.clone(),
            );
            driver.add_stimulus(stim);
        } else if ty.is_release() {
            // decaying exponential wave:
            let wave = Exp::new_at(amp, start, 0.125 * clock_phase_duration /* half-life */);
            let stim = CurlStimulus::new(
                region.clone(),
                wave.clone(),
            );
            driver.add_stimulus(stim);
        } // else: Float
        // sine wave (untested):
        // let wave = Sinusoid1::from_wavelength(direction as f32 * input_magnitude / area, clock_phase_duration * 2.0)
        //     .half_cycle()
        //     .shifted(clock_phase_duration * cycle as f32);
    };

    //////// define the control signals/transitions
    // each row N denotes the drive currents at clock cycle N.
    // each col M denotes the drive current at core M.
    use DriveType::*;
    // let s0_init_state = HoldHigh; // logic high
    let s0_init_state = HoldLow; // logic low
    let s0_rel_state = ReleaseLow;
    let drive_map = [
        // charge each device to '1'
        [s0_init_state,HoldHigh,     HoldHigh,     HoldHigh,     HoldHigh],
        // this is when we'd ordinarily open S0 for write
        [s0_rel_state, HoldHigh,     HoldHigh,     HoldHigh,     HoldHigh],

        // this is when S0 would ordinarily receive a write
        [Float,        HoldHigh,     HoldHigh,     HoldHigh,     HoldHigh],
        // open S1 for write
        [Float,        ReleaseHigh,  HoldHigh,     HoldHigh,     HoldHigh],

        // write S0' => S1
        [HoldLow,      Float,        HoldHigh,     HoldHigh,     HoldHigh],
        // open S2 for write
        [HoldLow,      Float,        ReleaseHigh,  HoldHigh,     HoldHigh],

        // write S1' => S2
        [HoldLow,      HoldLow,      Float,        HoldHigh,     HoldHigh],
        // open S3 for write
        [HoldLow,      HoldLow,      Float,        ReleaseHigh,  HoldHigh],

        // write S2' => S3; RESET S0
        [HoldHigh,     HoldLow,      HoldLow,      Float,        HoldHigh],
        // open S4 for write
        [HoldHigh,     HoldLow,      HoldLow,      Float,        ReleaseHigh],

        // write S3' => S4; RESET S1
        [HoldHigh,     HoldHigh,     HoldLow,      HoldLow,      Float],
        // open S0 for write
        [ReleaseHigh,  HoldHigh,     HoldLow,      HoldLow,      Float],

        // write S4' => output; RESET S2
        [Float,        HoldHigh,     HoldHigh,     HoldLow,      HoldLow],
        // open S1 for write
        [Float,        ReleaseHigh,  HoldHigh,     HoldLow,      HoldLow],
    ];

    let wire_mat = IsomorphicConductor::new(1e6f32.cast::<R>());
    // let ferro_mat = wire_mat;
    let ferro_mat = Ferroxcube3R1MH::new();

    let num_cores = drive_map[0].len();
    let last_core = num_cores - 1;

    let mut driver = Driver::new(Sim::new(
        sim_bounds(num_cores).to_index(feat_size), feat_size,
    ));
    driver.add_classical_boundary_explicit::<R, _>(sim_padding);

    //////// create the wires and toroids
    driver.fill_region(&input0, wire_mat);
    driver.fill_region(&sense(last_core), wire_mat);
    for core in 0..num_cores {
        driver.fill_region(&s(core), ferro_mat);
        driver.fill_region(&ctl(core), wire_mat);
        if core != last_core {
            driver.fill_region(&coupling(core), wire_mat);
        }
    }

    //////// monitor some measurements
    // driver.add_measurement(meas::CurrentLoop::new("input0", input0.clone()));
    for core in 0..num_cores {
        driver.add_measurement(meas::CurrentLoop::new(
            &format!("drive{}", core),
            ctl(core),
        ));
    }
    for core in 0..num_cores {
        let name = format!("sense{}", core);
        if core != last_core {
            driver.add_measurement(meas::CurrentLoop::new(
                &name, coupling(core),
            ));
        } else {
            driver.add_measurement(meas::CurrentLoop::new(
                &name, sense(core),
            ));
        }
    }
    for core in 0..num_cores {
        driver.add_measurement(meas::MagneticLoop::new(
            &format!("state{}", core),
            s(core),
        ));
    }


    let cycles = drive_map.len() as u32;
    let duration = Seconds(clock_phase_duration * (cycles + 3) as f32);

    for cycle in 0..cycles {
        for core in 0..num_cores {
            // current polarity of the drive wires is inverted in this model v.s. the blog article,
            // so invert our currents in order to achieve the magnetic states of the article.
            let sig = drive_map[cycle as usize][core as usize];
            control_signal(&mut driver, &ctl(core), cycle, sig);
        }
    }

    let prefix = "out/applications/multi_core_inverter/11/";
    let _ = std::fs::create_dir_all(&prefix);
    driver.add_state_file(&*format!("{}state.bc", prefix), 6400);
    driver.add_serializer_renderer(&*format!("{}frame-", prefix), 3200, None);
    driver.add_csv_renderer(&*format!("{}meas-detailed.csv", prefix), 100, None);
    driver.add_csv_renderer(&*format!("{}meas.csv", prefix), 800, None);
    driver.set_steps_per_stim(100);

    driver.step_until(duration);
}
