use crate::geom::{Coord, Meters, OrdMeters};

use coremem_cross::vec::Vec3;

use rayon::prelude::*;
use serde::{Serialize, Deserialize};
use std::collections::BTreeSet;
use std::sync::Arc;

mod primitives;
pub use primitives::*;


pub trait Region: Send + Sync {
    fn contains(&self, p: Meters) -> bool;
}

/// some (volume) which has a tangent vector everywhere inside/on it.
/// for example, a cylinder has tangents everywhere except its axis.
/// the returned vector should represent the area of the cross section.
pub trait HasCrossSection {
    fn cross_section_normal(&self, p: Meters) -> Vec3<f32>;
}

pub fn and<T1: Region + 'static, T2: Region + 'static>(r1: T1, r2: T2) -> Intersection<T1, T2> {
    Intersection::new2(r1, r2)
}

pub fn and_not<T1: Region + 'static, T2: Region + 'static>(r1: T1, r2: T2) -> Intersection<T1, InvertedRegion<T2>> {
    and(r1, InvertedRegion::new(r2))
}

pub fn union<T1: Region + 'static, T2: Region + 'static>(r1: T1, r2: T2) -> Union<T1, T2> {
    Union::new2(r1, r2)
}

/// returns true if there's a path (via the cardinal directions) from p0 to p1 within this region.
pub fn is_connected<R: Region, C: Coord>(r: &R, p0: C, p1: C, feat_size: f32) -> bool {
    distance_to(r, p0, p1, feat_size).is_some()
}

/// returns the minimal distance to crawl from p0 to p1 (via movements along a cardinal direction)
pub fn distance_to<R: Region, C: Coord>(r: &R, p0: C, p1: C, feat_size: f32) -> Option<f32> {
    // breadth-first crawl, where `visited` contains everything at an earlier depth
    // and `active_layer` contains nodes which are a minimal of `active_hops` from p0.
    let mut visited = BTreeSet::new();
    let mut active_layer = BTreeSet::new();
    active_layer.insert(p0.to_index(feat_size));
    let mut active_hops = 0;

    let p1idx = p1.to_index(feat_size);

    while !active_layer.is_empty() {
        // we have to mark these as visited *before* crawling them,
        // else neighbors might be crawled twice.
        visited.append(&mut active_layer.clone());

        if visited.contains(&p1idx) {
            return Some(active_hops as f32 * feat_size);
        }

        // add the neighbors to the list of places to crawl
        active_layer = active_layer
            .into_par_iter()
            .flat_map(|loc| loc.cardinal_neighbors())
            .filter(|neighbor| {
                !visited.contains(&neighbor) && r.contains(neighbor.to_meters(feat_size))
            })
            .collect();
        active_hops += 1;
    }
    None // not found
}

/// Region describing the entire simulation space
#[derive(Copy, Clone, Default, Serialize, Deserialize)]
pub struct WorldRegion;

impl Region for WorldRegion {
    fn contains(&self, _: Meters) -> bool {
        true
    }
}

#[derive(Clone, Default, Serialize, Deserialize)]
pub struct InvertedRegion<R>(R);

impl<R> InvertedRegion<R> {
    pub fn new(r: R) -> Self {
        Self(r)
    }
}

impl<R: Region> Region for InvertedRegion<R> {
    fn contains(&self, p: Meters) -> bool {
        !self.0.contains(p)
    }
}

#[derive(Clone, Default, Serialize, Deserialize)]
pub struct Union<R1, R2>(R1, R2);

pub type Union3<R1, R2, R3> = Union<Union<R1, R2>, R3>;
pub type Union4<R1, R2, R3, R4> = Union<Union3<R1, R2, R3>, R4>;

impl<R1, R2> Union<R1, R2> {
    pub fn with<R: Region>(self, r: R) -> Union<Self, R> {
        Union::new2(self, r)
    }
    pub fn new2(r1: R1, r2: R2) -> Self {
        Self(r1, r2)
    }
    pub fn new3<R3: Region>(r1: R1, r2: R2, r3: R3) -> Union<Self, R3> {
        Union::new2(r1, r2).with(r3)
    }
    pub fn new4<R3: Region, R4: Region>(r1: R1, r2: R2, r3: R3, r4: R4) -> Union<Union<Self, R3>, R4> {
        Union::new2(r1, r2).with(r3).with(r4)
    }
}

impl<R1: Region, R2: Region> Region for Union<R1, R2> {
    fn contains(&self, p: Meters) -> bool {
        self.0.contains(p) || self.1.contains(p)
    }
}

#[derive(Clone, Default, Serialize, Deserialize)]
pub struct Intersection<R1, R2>(R1, R2);


impl<R1, R2> Intersection<R1, R2> {
    pub fn and<R3: Region>(self, r: R3) -> Intersection<Self, R3> {
        Intersection::new2(self, r)
    }
    pub fn new2(r1: R1, r2: R2) -> Self {
        Self(r1, r2)
    }
}

impl<R1: Region, R2: Region> Region for Intersection<R1, R2> {
    fn contains(&self, p: Meters) -> bool {
        self.0.contains(p) && self.1.contains(p)
    }
}

#[derive(Clone, Default, Serialize, Deserialize)]
pub struct Translate<R> {
    inner: R,
    shift: Meters,
}

impl<R> Translate<R> {
    pub fn new(inner: R, shift: Meters) -> Self {
        Self { inner, shift }
    }
}

impl<R: Region> Region for Translate<R> {
    fn contains(&self, p: Meters) -> bool {
        self.inner.contains(p - self.shift)
    }
}

#[derive(Clone, Default, Serialize, Deserialize)]
pub struct SwapXZ<R> {
    inner: R,
}

impl<R> SwapXZ<R> {
    pub fn new(inner: R) -> Self {
        Self { inner }
    }
}

impl<R: Region> Region for SwapXZ<R> {
    fn contains(&self, p: Meters) -> bool {
        let p = Meters::new(p.z(), p.y(), p.z());
        self.inner.contains(p)
    }
}

#[derive(Clone, Default, Serialize, Deserialize)]
pub struct SwapYZ<R> {
    inner: R,
}

impl<R> SwapYZ<R> {
    pub fn new(inner: R) -> Self {
        Self { inner }
    }
}

impl<R: Region> Region for SwapYZ<R> {
    fn contains(&self, p: Meters) -> bool {
        let mapped = Meters::new(p.x(), p.z(), p.y());
        self.inner.contains(mapped)
    }
}

/// "wraps" the inner region into a donut shape.
/// this donut is traced out over the xy plane, leaving z untouched.
/// the center of the donut is (0, 0).
/// the resulting region is mapped onto the original region y=[0, y_max]. x is just the radius
/// so that (0, 0) is mapped to (0, 0), and (1, 0) is mapped to (1, 0) and (0, 1) is mapped to
/// (1, 0.5*y_max) and (-5, 0) is mapped to (5, 0.5*y_max).
#[derive(Clone, Default, Serialize, Deserialize)]
pub struct Wrap<R> {
    inner: R,
    y_max: f32,
    about: Meters,
}

impl<R> Wrap<R> {
    pub fn new(inner: R, y_max: f32) -> Self {
        Self::new_about(inner, y_max, Meters::new(0.0, 0.0, 0.0))
    }

    pub fn new_about(inner: R, y_max: f32, about: Meters) -> Self {
        Self { inner, y_max, about }
    }

    fn map(&self, p: Meters) -> Meters {
        let p_rel = p - self.about;
        let xy = p_rel.xy();
        let arg = xy.y().atan2(xy.x());
        let rev = arg / std::f32::consts::PI * self.y_max;
        Meters::new(xy.mag() + self.about.x(), rev, p.z())
    }
}

impl<R: Region> Region for Wrap<R> {
    fn contains(&self, p: Meters) -> bool {
        self.inner.contains(self.map(p))
    }
}

#[derive(Clone, Default, Serialize, Deserialize)]
pub struct Dilate<R> {
    inner: R,
    rad: f32,
    res: f32,
}

impl<R> Dilate<R> {
    pub fn new(inner: R, rad: f32, res: f32) -> Self {
        Self { inner, rad, res }
    }
}

impl<R: Region> Region for Dilate<R> {
    fn contains(&self, p: Meters) -> bool {
        let rad_iters = (self.rad / self.res).ceil() as i32;
        let rad_range = -rad_iters..=rad_iters;
        let rad_sq = self.rad * self.rad;
        for zoff in rad_range.clone() {
            for yoff in rad_range.clone() {
                for xoff in rad_range.clone() {
                    let off = Meters::new(
                        xoff as f32 * self.res,
                        yoff as f32 * self.res,
                        zoff as f32 * self.res,
                    );
                    if off.mag_sq() <= rad_sq && self.inner.contains(p + off) {
                        return true;
                    }
                }
            }
        }
        false
    }
}

#[derive(Clone, Default, Serialize, Deserialize)]
pub struct Memoize<R> {
    #[serde(skip)]
    lut: Arc<dashmap::DashMap<OrdMeters, bool>>,
    inner: R,
}

impl<R> Memoize<R> {
    pub fn new(inner: R) -> Self {
        Self {
            lut: Arc::new(dashmap::DashMap::new()),
            inner,
        }
    }
}

impl<R: Region> Region for Memoize<R> {
    fn contains(&self, p: Meters) -> bool {
        *self.lut.entry(OrdMeters(p)).or_insert_with(|| self.inner.contains(p))
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use float_eq::assert_float_eq;

    fn assert_map<R>(w: &Wrap<R>, from: Meters, to: Meters) {
        let mapped = w.map(from);
        assert_float_eq!(mapped.x(), to.x(), abs <= 0.01);
        assert_float_eq!(mapped.y(), to.y(), abs <= 0.01);
        assert_float_eq!(mapped.z(), to.z(), abs <= 0.01);
    }

    #[test]
    fn wrap_basics() {
        let wrap = Wrap::new(Cube::default(), 10.0);

        // magnitude
        assert_map(&wrap, Meters::new(0.0, 0.0, 0.0), Meters::new(0.0, 0.0, 0.0));
        assert_map(&wrap, Meters::new(1.0, 0.0, 0.0), Meters::new(1.0, 0.0, 0.0));
        assert_map(&wrap, Meters::new(5.0, 0.0, 0.0), Meters::new(5.0, 0.0, 0.0));

        // angle
        assert_map(&wrap, Meters::new(0.0, 4.0, 0.0), Meters::new(4.0, 5.0, 0.0));
        assert_map(&wrap, Meters::new(0.0, -0.5, 0.0), Meters::new(0.5, -5.0, 0.0));
        assert_map(&wrap, Meters::new(1.0, 1.0, 0.0), Meters::new(1.41, 2.5, 0.0));
        assert_map(&wrap, Meters::new(-2.0, 0.0, 0.0), Meters::new(2.0, 10.0, 0.0));

        // z
        assert_map(&wrap, Meters::new(0.0, 0.0, 3.0), Meters::new(0.0, 0.0, 3.0));
        assert_map(&wrap, Meters::new(1.0, 1.0, -5.0), Meters::new(1.41, 2.5, -5.0));

    }

    #[test]
    fn wrap_about() {
        let wrap = Wrap::new_about(
            Cube::new(Meters::new(0.0, 0.0, 0.0), Meters::new(5.0, 5.0, 5.0)),
            10.0,
            Meters::new(-10.0, 0.0, -1000.0 /* irrelevant */),
        );

        assert_map(&wrap, Meters::new(1.0, 0.0, 0.0), Meters::new(1.0, 0.0, 0.0));
        assert_map(&wrap, Meters::new(-10.0, 10.0, 0.0), Meters::new(0.0, 5.0, 0.0));
        assert_map(&wrap, Meters::new(-20.0, 0.0, 0.0), Meters::new(0.0, 10.0, 0.0));

        // verify resulting x component
        assert_map(&wrap, Meters::new(-21.0, 0.0, 0.0), Meters::new(1.0, 10.0, 0.0));
        assert_map(&wrap, Meters::new(-19.0, 0.0, 0.0), Meters::new(-1.0, 10.0, 0.0));
        assert_map(&wrap, Meters::new(-10.0, 9.0, 0.0), Meters::new(-1.0, 5.0, 0.0));
        assert_map(&wrap, Meters::new(-10.0, -9.0, 0.0), Meters::new(-1.0, -5.0, 0.0));

        // z should be no-op
        assert_map(&wrap, Meters::new(-10.0, -9.0, 5.0), Meters::new(-1.0, -5.0, 5.0));
    }
}
